// Copyright 2019, the Chromium project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

import 'dart:async';
import 'dart:convert';
import 'package:flutter/foundation.dart' show kIsWeb;
import 'package:flutter_driver/driver_extension.dart';
import 'package:flutter_test/flutter_test.dart' as flutter_test;
import 'package:flutter_test/flutter_test.dart' show Timeout;
// ignore: implementation_imports
import 'package:test_api/src/backend/invoker.dart' show Invoker;
// ignore: implementation_imports
import 'package:test_api/src/backend/group.dart' show Group;
// ignore: implementation_imports
import 'package:test_api/src/backend/group_entry.dart' show GroupEntry;

import 'common.dart';

export 'package:flutter_test/flutter_test.dart';

typedef TestMain = void Function();

String _currentTestId = '-1';
List<Map<String, dynamic>> _pendingEvents = [];
Map<Object, Map<String, String?>> _testingTests = {};
Map<Object, Map<String, String?>> _testingGroups = {};

void _iterateGroupEntries(GroupEntry groupEntry, GroupEntry parentEntry) {
  if (groupEntry is Group) {
    if (_testingGroups[groupEntry] == null) {
      Map<String, String?>? parentGroup = _testingGroups[parentEntry];
      _testingGroups[groupEntry] = {
        kId: (_testingGroups.length + 1).toString(),
        kName: groupEntry.name.replaceFirst(parentEntry.name, '').trim(),
        "parentGroupId": parentGroup != null ? parentGroup[kId] : null,
        "parentGroupName": parentGroup != null ? parentGroup[kName] : null,
      };
    }
    groupEntry.entries.forEach((entry) {
      _iterateGroupEntries(entry, groupEntry);
    });
  } else {
    if (_testingTests[groupEntry] == null) {
      Map<String, String?> parentGroup = _testingGroups[parentEntry]!;
      _testingTests[groupEntry] = {
        kId: (_testingTests.length + 1).toString(),
        kName: groupEntry.name.replaceFirst(parentEntry.name, '').trim(),
        kGroupId: parentGroup[kId],
        kGroupName: parentGroup[kName],
      };
    }
  }
}

final _zoneSpecification = ZoneSpecification(
    print: (Zone self, ZoneDelegate parent, Zone zone, String message) {
  if (kIsWeb) {
    // TODO(salakar) skipping web stack trace logging for now until source-map support added.
    if (message.contains('main.dart.js')) {
      return;
    }
    // TODO(salakar) skipping reporter log lines for web
    if (message.startsWith(r'[0-9]{2}:[0-9]{2}\s\+.*')) {
      return;
    }
  } else {
    // Skipping test status output since we make our own.
    String stacktrace = StackTrace.current.toString();
    if (stacktrace.contains('_Reporter._progressLine') ||
        stacktrace.contains('_Reporter._onError')) {
      return;
    }
  }
  _pendingEvents.add({
    kAction: EventActionPrint,
    kData: {
      kMessage: message,
      kTestId: _currentTestId,
    },
  });
});

dynamic Function() _zonedLifecycleBody(dynamic Function() body) {
  return () async {
    return runZonedGuarded(() async {
      await body.call();
    }, (e, stacktrace) {
      _pendingEvents.add({
        kAction: EventActionError,
        kData: {
          kError: e.toString(),
          kStacktrace: stacktrace.toString(),
        },
      });
    }, zoneSpecification: _zoneSpecification);
  };
}

void group(Object description, void Function() body, {dynamic skip}) {
  flutter_test.group(description, body, skip: skip);
}

void setUp(dynamic Function() body) {
  flutter_test.setUp(_zonedLifecycleBody(body));
}

void tearDown(dynamic Function() body) {
  flutter_test.tearDown(_zonedLifecycleBody(body));
}

void setUpAll(dynamic Function() body) {
  flutter_test.setUpAll(_zonedLifecycleBody(body));
}

void tearDownAll(dynamic Function() body) {
  flutter_test.tearDownAll(_zonedLifecycleBody(body));
}

void test(
  Object description,
  dynamic Function() body, {
  String? testOn,
  Timeout? timeout,
  dynamic skip,
  dynamic tags,
  Map<String, dynamic>? onPlatform,
  int? retry,
}) {
  flutter_test.test(
    description,
    body,
    testOn: testOn,
    timeout: timeout,
    skip: skip,
    onPlatform: onPlatform,
    tags: tags,
    retry: retry,
  );
}

void main(testMain) {
  enableFlutterDriverExtension(handler: (data) async {
    if (_pendingEvents.isEmpty) return '';
    return json.encode(_pendingEvents.removeAt(0));
  });

  runZonedGuarded(() {
    flutter_test.setUpAll(() {
      Invoker.current!.liveTest.groups.forEach((group) {
        if (_testingGroups[group] == null) {
          _testingGroups[group] = {
            kId: (_testingGroups.length + 1).toString(),
            kName: group.name,
          };
        }
        group.entries.forEach((entry) {
          _iterateGroupEntries(entry, group);
        });
      });
      _pendingEvents.add({
        kAction: EventActionSuiteStart,
        kData: {
          kTests: _testingTests.values.toList(),
          kGroups: _testingGroups.values.toList(),
          kWeb: kIsWeb,
        }
      });
    });

    flutter_test.tearDownAll(() => _pendingEvents.add({
          kAction: EventActionSuiteComplete,
          kData: {},
        }));

    flutter_test.setUp(() {
      Map<String, String?> currentTest =
          _testingTests[Invoker.current!.liveTest.test]!;
      _currentTestId = currentTest[kId]!;
      _pendingEvents.add({
        kAction: EventActionTestStart,
        kData: {
          kId: _currentTestId,
        }
      });
    });

    flutter_test.tearDown(() {
      final currentLiveTest = Invoker.current!.liveTest;
      Map eventData = {
        kId: _currentTestId.toString(),
        'result': currentLiveTest.state.result.toString(),
        kPassed: currentLiveTest.state.result.isPassing,
      };
      if (eventData[kPassed] == false) {
        AsyncError errorForTest =
            currentLiveTest.errors[currentLiveTest.errors.length - 1];
        eventData[kError] = errorForTest.error.toString();
        eventData[kStacktrace] = errorForTest.stackTrace.toString();
      }
      _pendingEvents.add({
        kAction: EventActionTestComplete,
        kData: eventData,
      });
    });

    testMain();
  }, (e, stacktrace) {
    _pendingEvents.add({
      kAction: EventActionError,
      kData: {
        kError: e.toString(),
        kStacktrace: stacktrace.toString(),
        kWeb: kIsWeb,
      },
    });
    _pendingEvents.add({
      kAction: EventActionSuiteComplete,
      kData: {},
    });
  }, zoneSpecification: _zoneSpecification);
}
